@@ -95,6 +95,14 @@ struct _GtkFileSystemGnomeVFS
 
   char *desktop_uri;
   char *home_uri;
+
+  /* For /afs and /net */
+  struct stat afs_statbuf;
+  struct stat net_statbuf;
+
+  guint have_afs : 1;
+  guint have_net : 1;
+
   guint locale_encoded_filenames : 1;
 };
 
@@ -125,7 +133,9 @@ struct _GtkFileFolderGnomeVFS
 
   GtkFileSystemGnomeVFS *system;
 
-  GHashTable *children;
+  GHashTable *children; /* NULL if destroyed */
+
+  guint is_afs_or_net : 1;
 };
 
 typedef struct _FolderChild FolderChild;
@@ -134,8 +144,21 @@ struct _FolderChild
 {
   gchar *uri; /* canonical */
   GnomeVFSFileInfo *info;
+  guint reloaded : 1; /* used when reloading a folder */
 };
 
+/* When getting a folder, we launch two asynchronous processes, the monitor for
+ * changes and the async directory load.  Unfortunately, the monitor will feed
+ * us all the filenames.  So, to avoid duplicate emissions of the "files-added"
+ * signal, we keep a "reloaded" flag in each FolderChild structure.
+ *
+ * If we re-get a live folder, we don't re-create its monitor, but we do restart
+ * its async loader.  When we (re)read the information for a file, we turn on
+ * its "reloaded" flag.  When the async load terminates, we purge out all the
+ * structures which don't have the flag set, and emit "files-removed" signals
+ * for them.
+ */
+
 #ifdef USE_GCONF
 /* GConf paths for the bookmarks; keep these two in sync */
 #define BOOKMARKS_KEY  "/desktop/gnome/interface/bookmark_folders"
@@ -220,6 +243,7 @@ static void  gtk_file_folder_gnome_vfs_class_init (GtkFileFolderGnomeVFSClass *c
 static void  gtk_file_folder_gnome_vfs_iface_init (GtkFileFolderIface         *iface);
 static void  gtk_file_folder_gnome_vfs_init       (GtkFileFolderGnomeVFS      *impl);
 static void  gtk_file_folder_gnome_vfs_finalize   (GObject                    *object);
+static void  gtk_file_folder_gnome_vfs_dispose    (GObject                    *object);
 
 static GtkFileInfo *gtk_file_folder_gnome_vfs_get_info      (GtkFileFolder      *folder,
 							     const GtkFilePath  *path,
@@ -264,8 +288,13 @@ static void client_notify_cb (GConfClient *client,
 			      gpointer     data);
 #endif
 
+static gchar *make_child_uri (const gchar *base_uri,
+			      const gchar *child_name,
+			      GError     **error);
+
 static FolderChild *folder_child_new (const char       *uri,
-				      GnomeVFSFileInfo *info);
+				      GnomeVFSFileInfo *info,
+				      gboolean          reloaded);
 static void         folder_child_free (FolderChild *child);
 
 static void     set_vfs_error     (GnomeVFSResult   result,
@@ -358,6 +387,10 @@ gtk_file_system_gnome_vfs_init (GtkFileSystemGnomeVFS *system_vfs)
 #endif
 
 
+  bindtextdomain (GETTEXT_PACKAGE, GNOMEUILOCALEDIR);
+#ifdef HAVE_BIND_TEXTDOMAIN_CODESET 
+  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
+#endif
   name = g_build_filename (g_get_home_dir (), "Desktop", NULL);
   system_vfs->desktop_uri = (char *)gtk_file_system_filename_to_path (GTK_FILE_SYSTEM (system_vfs),
 								      name);
@@ -366,7 +399,7 @@ gtk_file_system_gnome_vfs_init (GtkFileSystemGnomeVFS *system_vfs)
 								   g_get_home_dir ());
   
   system_vfs->locale_encoded_filenames = (getenv ("G_BROKEN_FILENAMES") != NULL);
-  system_vfs->folders = g_hash_table_new (g_str_hash, g_str_equal);
+  system_vfs->folders = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
 
 #ifdef USE_GCONF
   system_vfs->client = gconf_client_get_default ();
@@ -402,6 +435,25 @@ gtk_file_system_gnome_vfs_init (GtkFileSystemGnomeVFS *system_vfs)
     g_signal_connect_object (system_vfs->volume_monitor, "drive-disconnected",
 			     G_CALLBACK (drive_connect_disconnect_cb), system_vfs, 0);
 
+  /* Check for AFS */
+
+  if (stat ("/afs", &system_vfs->afs_statbuf) == 0)
+    system_vfs->have_afs = TRUE;
+  else
+    system_vfs->have_afs = FALSE;
+
+  if (stat ("/net", &system_vfs->net_statbuf) == 0)
+    system_vfs->have_net = TRUE;
+  else
+    system_vfs->have_net = FALSE;
+}
+
+static void
+unref_folder (gpointer key, 
+	      gpointer value,
+	      gpointer data)
+{
+  g_object_unref (G_OBJECT (value));
 }
 
 static void
@@ -411,6 +463,7 @@ gtk_file_system_gnome_vfs_finalize (GObject *object)
 
   g_free (system_vfs->desktop_uri);
   g_free (system_vfs->home_uri);
+  g_hash_table_foreach (system_vfs->folders, unref_folder, NULL);
   g_hash_table_destroy (system_vfs->folders);
 
 #ifdef USE_GCONF
@@ -490,7 +543,7 @@ gtk_file_system_gnome_vfs_list_volumes (GtkFileSystem *file_system)
   volume = gnome_vfs_volume_monitor_get_volume_for_path (system_vfs->volume_monitor, "/");
   if (volume)
     result = g_slist_prepend (result, volume);
-  
+
   return result;
 }
 
@@ -535,24 +588,29 @@ gtk_file_system_gnome_vfs_get_volume_for_path (GtkFileSystem     *file_system,
 }
 
 
-static gboolean
-remove_all  (gpointer  key,
-	     gpointer  value,
-	     gpointer  user_data)
+static void
+unmark_all_fn  (gpointer  key,
+		gpointer  value,
+		gpointer  user_data)
 {
-  return TRUE;
+  FolderChild *child;
+
+  child = value;
+  child->reloaded = FALSE;
 }
 
 static void
 load_dir (GtkFileFolderGnomeVFS *folder_vfs)
 {
   int num_items;
+  GnomeVFSFileInfoOptions vfs_options;
 
   if (folder_vfs->async_handle)
-    gnome_vfs_async_cancel (folder_vfs->async_handle);
+    {
+      gnome_vfs_async_cancel (folder_vfs->async_handle);
+      g_hash_table_foreach (folder_vfs->children, unmark_all_fn, NULL);
+    }
 
-  g_hash_table_foreach_remove (folder_vfs->children,
-			       remove_all, NULL);
   gnome_authentication_manager_push_async ();
 
   if (g_str_has_prefix (folder_vfs->uri, "file:"))
@@ -560,6 +618,11 @@ load_dir (GtkFileFolderGnomeVFS *folder_vfs)
   else
     num_items = ITEMS_PER_REMOTE_NOTIFICATION;
 
+  if (folder_vfs->is_afs_or_net)
+    vfs_options = GNOME_VFS_FILE_INFO_DEFAULT;
+  else
+    vfs_options = get_options (folder_vfs->types);
+
   gnome_vfs_async_load_directory (&folder_vfs->async_handle,
 				  folder_vfs->uri,
 				  get_options (folder_vfs->types),
@@ -569,6 +632,89 @@ load_dir (GtkFileFolderGnomeVFS *folder_vfs)
   gnome_authentication_manager_pop_async ();
 }
 
+static GnomeVFSFileInfo *
+vfs_info_new_from_afs_or_net_folder (const char *basename)
+{
+  GnomeVFSFileInfo *vfs_info;
+
+  vfs_info = gnome_vfs_file_info_new ();
+
+  vfs_info->name = g_strdup (basename);
+  vfs_info->valid_fields = GNOME_VFS_FILE_INFO_FIELDS_TYPE | GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE;
+  vfs_info->type = GNOME_VFS_FILE_TYPE_DIRECTORY;
+  vfs_info->mime_type = g_strdup ("x-directory/normal");
+
+  return vfs_info;
+}
+
+static void
+load_afs_dir (GtkFileFolderGnomeVFS *folder_vfs)
+{
+  GDir *dir;
+  char *pathname;
+  char *hostname;
+  const char *basename;
+  GSList *added_uris;
+  GSList *changed_uris;
+
+  g_assert (folder_vfs->is_afs_or_net);
+
+  pathname = g_filename_from_uri (folder_vfs->uri, &hostname, NULL);
+  g_assert (pathname != NULL); /* Must be already valid */
+  g_assert (hostname == NULL); /* Must be a local path */
+
+  dir = g_dir_open (pathname, 0, NULL);
+  if (!dir)
+    return;
+
+  added_uris = changed_uris = NULL;
+
+  while ((basename = g_dir_read_name (dir)) != NULL)
+    {
+      char *uri;
+      FolderChild *child;
+      GnomeVFSFileInfo *vfs_info;
+
+      uri = make_child_uri (folder_vfs->uri, basename, NULL);
+      if (!uri)
+	continue;
+
+      vfs_info = vfs_info_new_from_afs_or_net_folder (basename);
+
+      child = g_hash_table_lookup (folder_vfs->children, uri);
+      if (child)
+	{
+	  gnome_vfs_file_info_unref (child->info);
+
+	  child->info = vfs_info;
+	  gnome_vfs_file_info_ref (vfs_info);
+
+	  changed_uris = g_slist_prepend (changed_uris, child->uri);
+	}
+      else
+	{
+	  child = folder_child_new (uri, vfs_info, FALSE);
+	  g_hash_table_insert (folder_vfs->children, child->uri, child);
+
+	  added_uris = g_slist_prepend (added_uris, child->uri);
+	}
+
+      gnome_vfs_file_info_unref (vfs_info);
+      g_free (uri);
+    }
+
+  if (added_uris)
+    {
+      g_signal_emit_by_name (folder_vfs, "files-added", added_uris);
+      g_slist_free (added_uris);
+    }
+  if (changed_uris)
+    {
+      g_signal_emit_by_name (folder_vfs, "files-changed", changed_uris);
+      g_slist_free (changed_uris);
+    }
+}
+
 static char *
 make_uri_canonical (const char *uri)
 {
@@ -668,7 +814,6 @@ gtk_file_system_gnome_vfs_get_folder (GtkFileSystem     *file_system,
     }
 
   vfs_info_type = vfs_info->type;
-  gnome_vfs_file_info_unref (vfs_info);
 
   if (vfs_info_type != GNOME_VFS_FILE_TYPE_DIRECTORY)
     {
@@ -678,24 +823,45 @@ gtk_file_system_gnome_vfs_get_folder (GtkFileSystem     *file_system,
 		   _("%s is not a folder"),
 		   uri);
       g_free (uri);
+      gnome_vfs_file_info_unref (vfs_info);
       return NULL;
     }
 
   folder_vfs = g_object_new (GTK_TYPE_FILE_FOLDER_GNOME_VFS, NULL);
 
-  gnome_authentication_manager_push_sync ();
+  /* Check for AFS */
+  if (GNOME_VFS_FILE_INFO_LOCAL (vfs_info)
+      && (vfs_info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_DEVICE) != 0
+      && (vfs_info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_INODE) != 0
+      && ((system_vfs->have_afs
+	   && system_vfs->afs_statbuf.st_dev == vfs_info->device
+	   && system_vfs->afs_statbuf.st_ino == vfs_info->inode)
+	  || (system_vfs->have_net
+	      && system_vfs->net_statbuf.st_dev == vfs_info->device
+	      && system_vfs->net_statbuf.st_ino == vfs_info->inode)))
+    folder_vfs->is_afs_or_net = TRUE;
+  else
+    folder_vfs->is_afs_or_net = FALSE;
+
+  gnome_vfs_file_info_unref (vfs_info);
+
   monitor = NULL;
-  result = gnome_vfs_monitor_add (&monitor,
-				  uri,
-				  GNOME_VFS_MONITOR_DIRECTORY,
-				  monitor_callback, folder_vfs);
-  gnome_authentication_manager_pop_sync ();
-  if (result != GNOME_VFS_OK && result != GNOME_VFS_ERROR_NOT_SUPPORTED)
+
+  if (!folder_vfs->is_afs_or_net)
     {
-      g_free (uri);
-      g_object_unref (folder_vfs);
-      set_vfs_error (result, uri, error);
-      return NULL;
+      gnome_authentication_manager_push_sync ();
+      result = gnome_vfs_monitor_add (&monitor,
+				      uri,
+				      GNOME_VFS_MONITOR_DIRECTORY,
+				      monitor_callback, folder_vfs);
+      gnome_authentication_manager_pop_sync ();
+      if (result != GNOME_VFS_OK && result != GNOME_VFS_ERROR_NOT_SUPPORTED)
+	{
+	  g_free (uri);
+	  g_object_unref (folder_vfs);
+	  set_vfs_error (result, uri, error);
+	  return NULL;
+	}
     }
 
   folder_vfs->system = system_vfs;
@@ -800,6 +966,8 @@ volume_mount_cb (gboolean succeeded,
 {
   struct mount_closure *closure;
 
+  gdk_threads_enter ();
+  
   closure = data;
 
   closure->succeeded = succeeded;
@@ -811,6 +979,8 @@ volume_mount_cb (gboolean succeeded,
     }
 
   g_main_loop_quit (closure->loop);
+
+  gdk_threads_leave ();
 }
 
 static gboolean
@@ -830,9 +1000,9 @@ gtk_file_system_gnome_vfs_volume_mount (GtkFileSystem        *file_system,
       gnome_vfs_drive_mount (GNOME_VFS_DRIVE (volume), volume_mount_cb, &closure);
       gnome_authentication_manager_pop_sync ();
 
-      GDK_THREADS_LEAVE ();
+      gdk_threads_leave ();
       g_main_loop_run (closure.loop);
-      GDK_THREADS_ENTER ();
+      gdk_threads_enter ();
       g_main_loop_unref (closure.loop);
 
       if (closure.succeeded)
@@ -1327,7 +1497,12 @@ static GtkFilePath *
 gtk_file_system_gnome_vfs_filename_to_path (GtkFileSystem *file_system,
 					    const gchar   *filename)
 {
-  gchar *uri = gnome_vfs_get_uri_from_local_path (filename);
+  gchar *uri;
+  
+  if (!filename [0])
+	  return NULL;
+
+  uri = gnome_vfs_get_uri_from_local_path (filename);
   if (uri)
     {
       gchar *canonical;
@@ -1389,7 +1564,7 @@ get_vfs_info (GtkFileSystem     *file_system,
     }
   
   gtk_file_path_free (parent_path);
-  
+
   return info;
   
 }
@@ -1611,10 +1786,19 @@ gtk_file_system_gnome_vfs_insert_bookmark (GtkFileSystem     *file_system,
 
   for (l = bookmarks; l; l = l->next)
     {
-      const char *bookmark;
+      char *bookmark, *space;
 
       bookmark = l->data;
-      if (strcmp (bookmark, uri) == 0)
+      
+      space = strchr (bookmark, ' ');
+      if (space)
+	*space = '\0';
+      if (strcmp (bookmark, uri) != 0)
+	{
+	  if (space)
+	    *space = ' ';
+	}
+      else
 	{
 	  g_set_error (error,
 		       GTK_FILE_SYSTEM_ERROR,
@@ -1659,10 +1843,19 @@ gtk_file_system_gnome_vfs_remove_bookmark (GtkFileSystem     *file_system,
 
   for (l = bookmarks; l; l = l->next)
     {
-      const char *bookmark;
+      char *bookmark, *space;
 
-      bookmark = l->data;
-      if (strcmp (bookmark, uri) == 0)
+      bookmark = (char *)l->data;
+      space = strchr (bookmark, ' ');
+      if (space)
+	*space = '\0';
+
+      if (strcmp (bookmark, uri) != 0)
+	{
+	  if (space)
+	    *space = ' ';
+	}
+      else
 	{
 	  g_free (l->data);
 	  bookmarks = g_slist_remove_link (bookmarks, l);
@@ -1706,10 +1899,14 @@ gtk_file_system_gnome_vfs_list_bookmarks (GtkFileSystem *file_system)
 
   for (l = bookmarks; l; l = l->next)
     {
-      const char *name;
-
-      name = l->data;
-      result = g_slist_prepend (result, gtk_file_system_uri_to_path (file_system, name));
+       char *bookmark, *space;
+  
+       bookmark = (char *)l->data;
+       space = strchr (bookmark, ' ');
+       if (space)
+	 *space = '\0';
+       
+       result = g_slist_prepend (result, gtk_file_system_uri_to_path (file_system, bookmark));
     }
 
   bookmark_list_free (bookmarks);
@@ -1839,6 +2036,7 @@ gtk_file_folder_gnome_vfs_class_init (GtkFileFolderGnomeVFSClass *class)
   folder_parent_class = g_type_class_peek_parent (class);
 
   gobject_class->finalize = gtk_file_folder_gnome_vfs_finalize;
+  gobject_class->dispose = gtk_file_folder_gnome_vfs_dispose;
 }
 
 static void
@@ -1854,23 +2052,53 @@ gtk_file_folder_gnome_vfs_init (GtkFileFolderGnomeVFS *impl)
 {
 }
 
+static gboolean
+unref_at_idle (GObject *object)
+{
+  g_object_unref (object);
+  return FALSE;
+}
+
 static void
-gtk_file_folder_gnome_vfs_finalize (GObject *object)
+gtk_file_folder_gnome_vfs_dispose (GObject *object)
 {
   GtkFileFolderGnomeVFS *folder_vfs = GTK_FILE_FOLDER_GNOME_VFS (object);
   GtkFileSystemGnomeVFS *system_vfs = folder_vfs->system;
+  gboolean was_destroyed;
 
+  was_destroyed = folder_vfs->children == NULL;
+  
   if (folder_vfs->uri)
     g_hash_table_remove (system_vfs->folders, folder_vfs->uri);
+  folder_vfs->uri = NULL;
+  
   if (folder_vfs->async_handle)
     gnome_vfs_async_cancel (folder_vfs->async_handle);
+  folder_vfs->async_handle = NULL;
+  
   if (folder_vfs->monitor)
     gnome_vfs_monitor_cancel (folder_vfs->monitor);
+  folder_vfs->monitor = NULL;
+  
   if (folder_vfs->children)
     g_hash_table_destroy (folder_vfs->children);
+  folder_vfs->children = NULL;
 
-  g_free (folder_vfs->uri);
+  if (!was_destroyed)
+    {
+      /* The folder is now marked as destroyed, we
+       * free it at idle time to avoid races with the
+       * cancelled async job. See gnome_vfs_async_cancel()
+       */
+      g_object_ref (object);
+      g_idle_add ((GSourceFunc)unref_at_idle, object);
+    }
+}
 
+
+static void
+gtk_file_folder_gnome_vfs_finalize (GObject *object)
+{
   folder_parent_class->finalize (object);
 }
 
@@ -1907,7 +2135,7 @@ lookup_folder_child_from_uri (GtkFileFolder *folder,
     {
       GSList *uris;
 
-      child = folder_child_new (uri, vfs_info);
+      child = folder_child_new (uri, vfs_info, folder_vfs->async_handle ? TRUE : FALSE);
       gnome_vfs_file_info_unref (vfs_info);
       g_hash_table_replace (folder_vfs->children, child->uri, child);
 
@@ -1998,14 +2226,15 @@ gtk_file_folder_gnome_vfs_list_children (GtkFileFolder  *folder,
 {
   GtkFileFolderGnomeVFS *folder_vfs = GTK_FILE_FOLDER_GNOME_VFS (folder);
 
-  load_dir (folder_vfs);
+  if (folder_vfs->is_afs_or_net)
+    load_afs_dir (folder_vfs);
+  else
+    load_dir (folder_vfs);
   
   *children = NULL;
 
   g_hash_table_foreach (folder_vfs->children, list_children_foreach, children);
 
-  *children = g_slist_reverse (*children);
-
   return TRUE;
 }
 
@@ -2091,13 +2320,15 @@ info_from_vfs_info (const gchar      *uri,
 
 static FolderChild *
 folder_child_new (const char       *uri,
-		  GnomeVFSFileInfo *info)
+		  GnomeVFSFileInfo *info,
+		  gboolean          reloaded)
 {
   FolderChild *child;
 
   child = g_new (FolderChild, 1);
   child->uri = g_strdup (uri);
   child->info = info;
+  child->reloaded = reloaded ? TRUE : FALSE;
   gnome_vfs_file_info_ref (child->info);
 
   return child;
@@ -2165,7 +2396,9 @@ volume_mount_unmount_cb (GnomeVFSVolumeMonitor *monitor,
 			 GnomeVFSVolume        *volume,
 			 GtkFileSystemGnomeVFS *system_vfs)
 {
+  gdk_threads_enter ();
   g_signal_emit_by_name (system_vfs, "volumes-changed");
+  gdk_threads_leave ();
 }
 
 /* Callback used when a drive gets connected or disconnected */
@@ -2174,7 +2407,72 @@ drive_connect_disconnect_cb (GnomeVFSVolumeMonitor *monitor,
 			     GnomeVFSDrive         *drive,
 			     GtkFileSystemGnomeVFS *system_vfs)
 {
+  gdk_threads_enter ();
   g_signal_emit_by_name (system_vfs, "volumes-changed");
+  gdk_threads_leave ();
+}
+
+struct purge_closure
+{
+  GSList *removed_uris;
+};
+
+/* Used from g_hash_table_foreach() in folder_purge_and_unmark() */
+static gboolean
+purge_fn (gpointer key,
+	  gpointer value,
+	  gpointer data)
+{
+  struct purge_closure *closure;
+  FolderChild *child;
+
+  child = value;
+  closure = data;
+
+  if (child->reloaded)
+    {
+      child->reloaded = FALSE;
+      return FALSE;
+    }
+  else
+    {
+      closure->removed_uris = g_slist_prepend (closure->removed_uris, child->uri);
+
+      if (child->info)
+	gnome_vfs_file_info_unref (child->info);
+
+      g_free (child);
+
+      return TRUE;
+    }
+}
+
+/* Purges out all the files that don't have the "reloaded" flag set */
+static void
+folder_purge_and_unmark (GtkFileFolderGnomeVFS *folder_vfs)
+{
+  struct purge_closure closure;
+
+  closure.removed_uris = NULL;
+  g_hash_table_foreach_steal (folder_vfs->children, purge_fn, folder_vfs);
+
+  if (closure.removed_uris)
+    {
+      GSList *l;
+
+      closure.removed_uris = g_slist_reverse (closure.removed_uris);
+      g_signal_emit_by_name (folder_vfs, "files-removed", closure.removed_uris);
+      
+      for (l = closure.removed_uris; l; l = l->next)
+	{
+	  char *uri;
+
+	  uri = l->data;
+	  g_free (uri);
+	}
+
+      g_slist_free (closure.removed_uris);
+    }
 }
 
 static void
@@ -2189,11 +2487,20 @@ directory_load_callback (GnomeVFSAsyncHandle *handle,
   GSList *added_uris = NULL;
   GSList *changed_uris = NULL;
 
+  /* This is called from an idle, we need to protect is as such */
+  gdk_threads_enter ();
+
+  /* See if this folder is already destroyed.
+   * This can happen if we cancel in another thread while
+   * directory_load_callback is just being called
+   */
+  if (folder_vfs->children == NULL)
+    return;
+  
   for (tmp_list = list; tmp_list; tmp_list = tmp_list->next)
     {
       GnomeVFSFileInfo *vfs_info;
       gchar *uri;
-      gboolean new = FALSE;
 
       vfs_info = tmp_list->data;
       if (strcmp (vfs_info->name, ".") == 0 ||
@@ -2206,20 +2513,28 @@ directory_load_callback (GnomeVFSAsyncHandle *handle,
 	{
 	  FolderChild *child;
 
-	  child = folder_child_new (uri, vfs_info);
+	  child = g_hash_table_lookup (folder_vfs->children, uri);
+	  if (child)
+	    {
+	      child->reloaded = TRUE;
+
+	      if (child->info)
+		gnome_vfs_file_info_unref (child->info);
+
+	      child->info = vfs_info;
+	      gnome_vfs_file_info_ref (child->info);
+
+	      changed_uris = g_slist_prepend (changed_uris, child->uri);
+	    }
+	  else
+	    {
+	      child = folder_child_new (uri, vfs_info, TRUE);
+	      g_hash_table_insert (folder_vfs->children, child->uri, child);
+
+	      added_uris = g_slist_prepend (added_uris, child->uri);
+	    }
 
 	  g_free (uri);
-
-	  if (!g_hash_table_lookup (folder_vfs->children, child->uri))
-	    new = TRUE;
-
-	  g_hash_table_replace (folder_vfs->children,
-				child->uri, child);
-
-	  if (new)
-	    added_uris = g_slist_prepend (added_uris, child->uri);
-	  else
-	    changed_uris = g_slist_prepend (changed_uris, child->uri);
 	}
     }
 
@@ -2239,7 +2554,10 @@ directory_load_callback (GnomeVFSAsyncHandle *handle,
       folder_vfs->async_handle = NULL;
 
       g_signal_emit_by_name (folder_vfs, "finished-loading");
+      folder_purge_and_unmark (folder_vfs);
     }
+
+  gdk_threads_leave ();
 }
 
 static void
@@ -2252,6 +2570,13 @@ monitor_callback (GnomeVFSMonitorHandle   *handle,
   GtkFileFolderGnomeVFS *folder_vfs = user_data;
   GSList *uris;
 
+  /* This is called from an idle, we need to protect is as such */
+  gdk_threads_enter ();
+
+  /* Check if destroyed */
+  if (folder_vfs->children == NULL)
+    return;
+  
   switch (event_type)
     {
     case GNOME_VFS_MONITOR_EVENT_CHANGED:
@@ -2261,6 +2586,7 @@ monitor_callback (GnomeVFSMonitorHandle   *handle,
 	GnomeVFSFileInfo *vfs_info;
 
 	vfs_info = gnome_vfs_file_info_new ();
+
 	gnome_authentication_manager_push_sync ();
 	result = gnome_vfs_get_file_info (info_uri, vfs_info, get_options (folder_vfs->types));
 	gnome_authentication_manager_pop_sync ();
@@ -2269,11 +2595,24 @@ monitor_callback (GnomeVFSMonitorHandle   *handle,
 	  {
 	    FolderChild *child;
 
-	    child = folder_child_new (info_uri, vfs_info);
-	    gnome_vfs_file_info_unref (vfs_info);
+	    child = g_hash_table_lookup (folder_vfs->children, info_uri);
 
-	    g_hash_table_replace (folder_vfs->children,
-				  child->uri, child);
+	    if (child)
+	      {
+		if (folder_vfs->async_handle)
+		  child->reloaded = TRUE;
+
+		if (child->info)
+		  gnome_vfs_file_info_unref (child->info);
+
+		child->info = vfs_info;
+	      }
+	    else
+	      {
+		child = folder_child_new (info_uri, vfs_info, folder_vfs->async_handle ? TRUE : FALSE);
+		gnome_vfs_file_info_unref (vfs_info);
+		g_hash_table_insert (folder_vfs->children, child->uri, child);
+	      }
 
 	    uris = g_slist_prepend (NULL, (gchar *)info_uri);
 	    if (event_type == GNOME_VFS_MONITOR_EVENT_CHANGED)
@@ -2298,6 +2637,8 @@ monitor_callback (GnomeVFSMonitorHandle   *handle,
     case GNOME_VFS_MONITOR_EVENT_METADATA_CHANGED:
       break;
     }
+
+  gdk_threads_leave ();
 }
 
 static gboolean
