@@ -26,6 +26,11 @@
 
 #include <string.h>
 
+#if defined(__APPLE__) && defined(HAVE_NSGETENVIRON) && defined(HAVE_CRT_EXTERNS_H)
+# include <crt_externs.h>
+# define environ (*_NSGetEnviron())
+#endif
+
 extern char **environ;
 
 /**
